{% extends "base.j2" %}
{% block SETUP_BENTO_BASE_IMAGE %}
{{ super() }}
USER root

{% call common.RUN(__enable_buildkit__) -%} {{ common.mount_cache("/var/cache/yum") }} {% endcall -%} set -eux && \
    yum upgrade -y \
    && yum install -y ca-certificates curl gcc gcc-c++ make bash shadow-utils

{% if __options__system_packages is not none %}
# Install user-defined system package
{% call common.RUN(__enable_buildkit__) -%} {{ common.mount_cache("/var/cache/yum") }} {% endcall -%} set -eux && \
    yum install -y {{ __options__system_packages | join(' ') }}
{% endif -%}
{% endblock %}
